/*
 *    Copyright (C) 2002  Claudio Agosti - vecna@s0ftpj.org
 *                                         claudio@blackhats.it
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * key daemon, to read the password for umpf system.
 *
 * P.S. the key is scattered on memory for protect our system from direct 
 * memory attack how presented on rmfbd 
 * http://www.s0ftpj.org/bfi/dev/BFi11-dev-06
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>

#include "shared-func.h"

#define getrandom(min, max) ((rand() % (int)(((max)+1) - (min))) + (min))
#define MAXCONNMB        4 

static struct termios stored_settings;
extern int errno;
char pidfile[UNIX_PATH_MAX], umpf_unix[UNIX_PATH_MAX];
static int fd;

char *rand_malloc(unsigned int size)
	{
	void *foo, *ret;
	static int foospace =32;

	srand(time(NULL));

	foospace *=getrandom(2, 256);
	foospace %=256;

	foo =malloc(foospace);
	ret =malloc(size);

	if(foo ==NULL || ret ==NULL)
		{
		perror("malloc");
		exit(EXIT_FAILURE);
		}

	memset(ret, 0x00, size);
	free(foo);

	return ret;
	}

void echo_off(void)
	{
	struct termios new_settings;

	tcgetattr(0, &stored_settings);

	new_settings = stored_settings;
	new_settings.c_lflag &= (~ECHO);

	tcsetattr(0, TCSANOW, &new_settings);

	return;
	}

void echo_on(void)
	{
	tcsetattr(0, TCSANOW, &stored_settings);

	return;
	}

void unlink_pidf(int sig)
	{
	unlink(pidfile);
	close(fd);
	unlink(umpf_unix);
	echo_on();

	printf("received signal %d on pid %d, cleaning file and quitting\n",
			sig, getpid());

	exit(0);
	}

void set_sig_trap(void)
	{
	int sigcnt;

	for(sigcnt =0; sigcnt !=__SIGRTMAX; sigcnt++)
		signal(sigcnt, unlink_pidf);
	}

void get_key(char *destkeybuff)
	{
	int nbytes;
	char buff_1[KEYL], buff_2[KEYL];
		
	printf("insert umpf key:");
	fflush(stdout); 
	echo_off();
	nbytes =read(STDIN_FILENO, &buff_1, KEYL);
	buff_1[nbytes -1] =0x00;
	echo_on();
	printf("\n");

	printf("retype umpf key:");
	fflush(stdout);
	echo_off();
	nbytes =read(STDIN_FILENO, &buff_2, KEYL);
	buff_2[nbytes -1] =0x00;
	echo_on();
	printf("\n");

	if(strcmp(buff_1, buff_2))
		fatal(NOT_ERRNO, "keys differ!\n");

	memcpy((void *)destkeybuff, (void *)&buff_1, KEYL);

	memset(&buff_1, 0x00, KEYL);
	memset(&buff_2, 0x00, KEYL);
	}

int main(int argc, char **argv)
	{
	char **key, key_read[KEYL], *home, pidn[5];
	int i, keylen, pid;
	FILE *pidf;
	struct stat umpf, umpf_pid;
	struct sockaddr_un unix_sock;

	if((home =getenv("HOME")) ==NULL)
		fatal(NOT_ERRNO, "unable to read $HOME from enviroment\n");

	set_sig_trap();
	
	/* read key via getpass() two times */
	get_key((char *)&key_read);

	/* work for user and after work for security */
	umask(S_IWGRP|S_IWOTH);
	
	/* check pidfile */
	snprintf(pidfile, UNIX_PATH_MAX, "%s/bin/keyd.pid", home);

	if(argc > 1 && !strcmp(argv[1], "-f"))
		{
		if((pidf =fopen(pidfile, "r")) ==NULL)
			fatal(USE_ERRNO, "unable to open %s", pidfile);

		fgets(pidn, 5, pidf);
		kill(atoi(pidn), SIGKILL);

		fclose(pidf);
		
		if(unlink(pidfile))
			fatal(USE_ERRNO, "unable to unlink %s", pidfile);
		}
	
	if(!stat(pidfile, &umpf_pid))
		{

		pidf =fopen(pidfile, "r");

		fgets(pidn, 5, pidf);

		fatal(NOT_ERRNO,"keyd already running, pid stored on %s\n"
				"if not, force unlinking of file with "
				"option -f\n", pidn);
		fclose(pidf);
		}
	else 
		if(errno !=ENOENT)
			fatal(USE_ERRNO, "unable to stat file %s", pidfile);

	if((pidf =fopen(pidfile, "w+")) ==NULL)
		fatal(USE_ERRNO, "unalbe to open on write mode %s", pidfile);

	pid =getpid();
	pid++;
	/* below is called fork() */
	fprintf(pidf, "%d\n", pid);

	fclose(pidf);
	
	/* to keep the key scattered all over the memory */
	key =(char **)rand_malloc(strlen(key_read));
	keylen =strlen(key_read);
	
	for(i =0; i !=keylen; i++)
		{
		key[i] =rand_malloc(sizeof(char));
		memcpy(key[i], &key_read[i], sizeof(char));
		}

	memset((void *)&key_read, 0x00, KEYL);
	
	/* to make srvr system */
	snprintf(umpf_unix, UNIX_PATH_MAX, "%s/bin/umpf.unix", home);

	if((fd =socket(PF_UNIX, SOCK_STREAM, 0x0000)) ==-1)
		fatal(USE_ERRNO, "error on UNIX socket");

	if(stat(umpf_unix, &umpf) !=-1)
		unlink(umpf_unix);
	else
		if(errno !=ENOENT)
			fatal(USE_ERRNO, "%s", umpf_unix);

	unix_sock.sun_family =AF_UNIX;
	memcpy(&unix_sock.sun_path, &umpf_unix, UNIX_PATH_MAX);

	if(bind(fd, (struct sockaddr *)&unix_sock, sizeof(unix_sock)))
		fatal(USE_ERRNO, "on binding unix socket %s", unix_sock);

	if(fork())
		exit(0);
	
	if(listen(fd, MAXCONNMB) ==1)
		fatal(USE_ERRNO, "error in listen");

	while(1)
		{
		int c_fd;
		struct sockaddr_un c_sun;
		int sun_len =sizeof(c_sun);

		c_sun.sun_family =AF_UNIX;
		memcpy(&c_sun.sun_path, &umpf_unix, UNIX_PATH_MAX);

		c_fd =accept(fd, (struct sockaddr *)&c_sun, &sun_len);
		if(c_fd ==-1)
			fatal(USE_ERRNO, "error on accept");

		for(i =0; i !=keylen; i++)
			write(c_fd, (void *)key[i], sizeof(char));

		close(c_fd);
		}
	
	fflush(stdout);

	return 0;
	}
